@@ -32,7 +32,7 @@ import os
 
 def bool_env(varname: str, default: bool) -> bool:
   """Read an environment variable and interpret it as a boolean.
-  
+
   True values are (case insensitive): 'y', 'yes', 't', 'true', 'on', and '1';
   false values are 'n', 'no', 'f', 'false', 'off', and '0'.
   Args:
@@ -60,4 +60,4 @@ flax_filter_frames = bool_env('FLAX_FILTER_FRAMES', True)
 flax_profile = bool_env('FLAX_PROFILE', False)
 
 # Whether to use the lazy rng implementation
-flax_lazy_rng = bool_env('FLAX_LAZY_RNG', False)
+flax_lazy_rng = bool_env('FLAX_LAZY_RNG', True)
