@@ -31,26 +31,28 @@ import jax.numpy as jnp
 import numpy as np
 
 
+def _pmap_device_order():
+  # match the default device assignments used in pmap:
+  # for single-host, that's the XLA default device assignment
+  # for multi-host, it's the order of jax.local_devices()
+  if jax.process_count() == 1:
+    return [d for d in xb.get_backend().get_default_device_assignment(
+        jax.device_count()) if d.process_index == jax.process_index()]
+  else:
+    return jax.local_devices()
+
+
 def replicate(tree, devices=None):
   """Replicates arrays to multiple devices.
 
   Args:
     tree: a pytree containing the arrays that should be replicated.
     devices: the devices the data is replicated to
-      (default: `jax.local_devices()`).
+      (default: same order as expected by `jax.pmap()`).
   Returns:
     A new pytree containing the replicated arrays.
   """
-  if devices is None:
-    # match the default device assignments used in pmap:
-    # for single-host, that's the XLA default device assignment
-    # for multi-host, it's the order of jax.local_devices()
-    if jax.process_count() == 1:
-      devices = [d for d in xb.get_backend().get_default_device_assignment(
-          jax.device_count()) if d.process_index == jax.process_index()]
-    else:
-      devices = jax.local_devices()
-
+  devices = devices or _pmap_device_order()
   return jax.device_put_replicated(tree, devices)
 
 
@@ -130,12 +132,14 @@ def prefetch_to_device(iterator, size, devices=None):
 
     devices: the list of devices to which the arrays should be prefetched.
 
+      Defaults to the order of devices expected by `jax.pmap`.
+
   Yields:
     The original items from the iterator where each ndarray is now a sharded to
     the specified devices.
   """
   queue = collections.deque()
-  devices = devices or jax.local_devices()
+  devices = devices or _pmap_device_order()
 
   def _prefetch(xs):
     if hasattr(jax, "device_put_sharded"):  # jax>=0.2.0
