@@ -14,28 +14,26 @@
 
 """Recurrent neural network modules.
 
-THe RNNCell modules are designed to fit in with the scan function in JAX::
-
-  _, initial_params = LSTMCell.init(rng_1, time_series[0])
-  model = nn.Model(LSTMCell, initial_params)
-  carry = LSTMCell.initialize_carry(rng_2, (batch_size,), memory_size)
-  carry, y = jax.lax.scan(model, carry, time_series)
-
+THe RNNCell modules can be scanned using lifted transforms. For more information
+see: https://flax.readthedocs.io/en/latest/design_notes/lift.html.
 """
 
 import abc
-from functools import partial
-from typing import (Any, Callable, Iterable, Optional, Tuple, Union)
-
-from flax.linen.module import Module, compact
-from flax.linen.activation import sigmoid, tanh
-from flax.linen.initializers import orthogonal, zeros
-from flax.linen.linear import Conv, Dense, default_kernel_init
+from functools import partial   # pylint: disable=g-importing-member
+from typing import Any, Callable, Mapping, Optional, Sequence, Tuple, Union
 
+from flax.linen.activation import sigmoid
+from flax.linen.activation import tanh
+from flax.linen.initializers import orthogonal
+from flax.linen.initializers import zeros
+from flax.linen.linear import Conv
+from flax.linen.linear import default_kernel_init
+from flax.linen.linear import Dense
+from flax.linen.linear import PrecisionLike
+from flax.linen.module import compact
+from flax.linen.module import Module
 from jax import numpy as jnp
-from jax import lax
 from jax import random
-
 import numpy as np
 
 PRNGKey = Any
@@ -50,7 +48,7 @@ class RNNCellBase(Module):
   @staticmethod
   @abc.abstractmethod
   def initialize_carry(rng, batch_dims, size, init_fn=zeros):
-    """initialize the RNN cell carry.
+    """Initialize the RNN cell carry.
 
     Args:
       rng: random number generator passed to the init_fn.
@@ -60,12 +58,14 @@ class RNNCellBase(Module):
     Returns:
       An initialized carry for the given RNN cell.
     """
-    pass
+    raise NotImplementedError
 
 
 class LSTMCell(RNNCellBase):
   r"""LSTM cell.
-    the mathematical definition of the cell is as follows
+
+  The mathematical definition of the cell is as follows
+
   .. math::
       \begin{array}{ll}
       i = \sigma(W_{ii} x + W_{hi} h + b_{hi}) \\
@@ -75,6 +75,7 @@ class LSTMCell(RNNCellBase):
       c' = f * c + i * g \\
       h' = o * \tanh(c') \\
       \end{array}
+
   where x is the input, h is the output of the previous time step, and c is
   the memory.
 
@@ -137,7 +138,7 @@ class LSTMCell(RNNCellBase):
 
   @staticmethod
   def initialize_carry(rng, batch_dims, size, init_fn=zeros):
-    """initialize the RNN cell carry.
+    """Initialize the RNN cell carry.
 
     Args:
       rng: random number generator passed to the init_fn.
@@ -153,13 +154,13 @@ class LSTMCell(RNNCellBase):
 
 
 class DenseParams(Module):
-  """Dummy module for creating parameters matching `flax.nn.Dense`."""
+  """Dummy module for creating parameters matching `flax.deprecated.nn.Dense`."""
 
   features: int
   use_bias: bool = True
   dtype: Dtype = jnp.float32
   param_dtype: Dtype = jnp.float32
-  precision: Any = None
+  precision: PrecisionLike = None
   kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = default_kernel_init
   bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = zeros
 
@@ -167,8 +168,10 @@ class DenseParams(Module):
   def __call__(self, inputs: Array) -> Tuple[Array, Array]:
     k = self.param(
         'kernel', self.kernel_init, (inputs.shape[-1], self.features))
-    b = (self.param('bias', self.bias_init, (self.features,))
-        if self.use_bias else jnp.zeros((self.features,)))
+    if self.use_bias:
+      b = self.param('bias', self.bias_init, (self.features,))
+    else:
+      b = jnp.zeros((self.features,))
     return k, b
 
 
@@ -178,8 +181,11 @@ class OptimizedLSTMCell(RNNCellBase):
   The parameters are compatible with `LSTMCell`. Note that this cell is often
   faster than `LSTMCell` as long as the hidden size is roughly <= 2048 units.
 
-  The mathematical definition of the cell is the same as `LSTMCell` and as follows
+  The mathematical definition of the cell is the same as `LSTMCell` and as
+  follows
+
   .. math::
+
       \begin{array}{ll}
       i = \sigma(W_{ii} x + W_{hi} h + b_{hi}) \\
       f = \sigma(W_{if} x + W_{hf} h + b_{hf}) \\
@@ -188,10 +194,11 @@ class OptimizedLSTMCell(RNNCellBase):
       c' = f * c + i * g \\
       h' = o * \tanh(c') \\
       \end{array}
+
   where x is the input, h is the output of the previous time step, and c is
   the memory.
 
-  Args:
+  Attributes:
     gate_fn: activation function used for gates (default: sigmoid).
     activation_fn: activation function used for output and memory update
       (default: tanh).
@@ -203,8 +210,8 @@ class OptimizedLSTMCell(RNNCellBase):
     dtype: the dtype of the computation (default: float32).
     param_dtype: the dtype passed to parameter initializers (default: float32).
   """
-  gate_fn: Callable = sigmoid
-  activation_fn: Callable = tanh
+  gate_fn: Callable[..., Any] = sigmoid
+  activation_fn: Callable[..., Any] = tanh
   kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = default_kernel_init
   recurrent_kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = orthogonal()
   bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = zeros
@@ -229,12 +236,12 @@ class OptimizedLSTMCell(RNNCellBase):
     hidden_features = h.shape[-1]
     inputs = jnp.asarray(inputs, self.dtype)
 
-    def _concat_dense(inputs, params, use_bias=True):
-      """
-      Concatenates the individual kernels and biases, given in params, into a 
-      single kernel and single bias for efficiency before applying them using 
-      dot_general.
-      """
+    def _concat_dense(inputs: Array,
+                      params: Mapping[str, Tuple[Array, Array]],
+                      use_bias: bool = True) -> Array:
+      # Concatenates the individual kernels and biases, given in params, into a
+      # single kernel and single bias for efficiency before applying them using
+      # dot_general.
       kernels, biases = zip(*params.values())
       kernel = jnp.asarray(jnp.concatenate(kernels, axis=-1), self.dtype)
 
@@ -276,7 +283,7 @@ class OptimizedLSTMCell(RNNCellBase):
 
   @staticmethod
   def initialize_carry(rng, batch_dims, size, init_fn=zeros):
-    """initialize the RNN cell carry.
+    """Initialize the RNN cell carry.
 
     Args:
       rng: random number generator passed to the init_fn.
@@ -295,14 +302,17 @@ class OptimizedLSTMCell(RNNCellBase):
 class GRUCell(RNNCellBase):
   r"""GRU cell.
 
-  the mathematical definition of the cell is as follows
+  The mathematical definition of the cell is as follows
+
   .. math::
+
       \begin{array}{ll}
       r = \sigma(W_{ir} x + W_{hr} h + b_{hr}) \\
       z = \sigma(W_{iz} x + W_{hz} h + b_{hz}) \\
       n = \tanh(W_{in} x + b_{in} + r * (W_{hn} h + b_{hn})) \\
-      h' = (1 - z) * n + z * h
+      h' = (1 - z) * n + z * h \\
       \end{array}
+
   where x is the input and h, is the output of the previous time step.
 
   Attributes:
@@ -367,7 +377,7 @@ class GRUCell(RNNCellBase):
 
   @staticmethod
   def initialize_carry(rng, batch_dims, size, init_fn=zeros):
-    """initialize the RNN cell carry.
+    """Initialize the RNN cell carry.
 
     Args:
       rng: random number generator passed to the init_fn.
@@ -423,9 +433,9 @@ class ConvLSTM(RNNCellBase):
   """
 
   features: int
-  kernel_size: Iterable[int]
-  strides: Optional[Iterable[int]] = None
-  padding: Union[str, Iterable[Tuple[int, int]]] = 'SAME'
+  kernel_size: Sequence[int]
+  strides: Optional[Sequence[int]] = None
+  padding: Union[str, Sequence[Tuple[int, int]]] = 'SAME'
   use_bias: bool = True
   dtype: Dtype = jnp.float32
   param_dtype: Dtype = jnp.float32
@@ -472,7 +482,7 @@ class ConvLSTM(RNNCellBase):
 
   @staticmethod
   def initialize_carry(rng, batch_dims, size, init_fn=zeros):
-    """initialize the RNN cell carry.
+    """Initialize the RNN cell carry.
 
     Args:
       rng: random number generator passed to the init_fn.
